"""
Configuration for pytest call

Configuration for pytest call, define database connection

xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/02"
__license__ = "AGPLv3"

import pytest
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from sqlalchemy.orm import Session

from xplordb.create_db import CreateDatabase
from xplordb.import_data import ImportData
from xplordb.import_ddl import ImportDDL
from xplordb.sql_utils import connect
from xplordb.sqlalchemy.base import create_session


def pytest_addoption(parser):
    parser.addoption("--host", action="store",
                     default='localhost')
    parser.addoption("--user", action="store",
                     default='postgres')
    parser.addoption("--password", action="store")
    parser.addoption("--port", action="store", default=5432)
    parser.addoption("--db", action="store")


@pytest.fixture(scope="module")
def host(request):
    return request.config.getoption("--host")


@pytest.fixture(scope="module")
def user(request):
    return request.config.getoption("--user")


@pytest.fixture(scope="module")
def password(request):
    return request.config.getoption("--password")


@pytest.fixture(scope="module")
def db(request):
    return request.config.getoption("--db")


@pytest.fixture(scope="module")
def port(request):
    return request.config.getoption("--port")


@pytest.fixture()
def db_session(host: str, port: int, db: str, user: str, password: str):
    """
    Create a pyscog2 connection with specified parameters
    parameters are defined from pytest call option.

    For example pytest --host localhost --port 5432 --db xplordb_test --user postgres --password postgres

    (default  values are defined in conftest.py)

    :param host: host used for connection
    :param port: port used for connection
    :param db: database name used for connection
    :param user:user used for connection
    :param password:password used for connection
    """
    session = connect(
        database=db,
        host=host,
        port=port,
        user=user,
        password=password)
    yield session
    session.close()


@pytest.fixture()
def sqlalchemy_session(host: str, port: int, db: str, user: str, password: str):
    """
    Create a sql alchemy session with specified parameters
    parameters are defined from pytest call option.

    For example pytest --host localhost --port 5432 --db xplordb_test --user postgres --password postgres

    (default  values are defined in conftest.py)

    :param host: host used for connection
    :param port: port used for connection
    :param db: database name used for connection
    :param user:user used for connection
    :param password:password used for connection
    """
    session = create_session(host, port, db, user, password)

    yield session
    session.close()


@pytest.fixture()
def litedb_no_data(host, user, password, db, port):
    """
    pytest fixture to create a xplordb database with lite schema without sample data
    """
    _import_db(host, user, password, db, port, False, False)


@pytest.fixture()
def litedb_with_data(host, user, password, db, port):
    """
    pytest fixture to create a xplordb database with lite schema and sample data
    """
    _import_db(host, user, password, db, port, True, False)


@pytest.fixture()
def fulldb_with_data(host, user, password, db, port):
    """
    pytest fixture to create a xplordb database with full schema with sample data
    """
    _import_db(host, user, password, db, port, True, True)


@pytest.fixture()
def fulldb_no_data(host, user, password, db, port):
    """
    pytest fixture to create a xplordb database with full schema without sample data
    """
    _import_db(host, user, password, db, port, False, True)


def _import_db(host: str, user: str, password: str, db: str, port: int,
               import_data: bool, full_db: bool):
    import_ddl_ = ImportDDL(host=host, user=user, password=password, db=db, port=port)
    import_ddl_.import_xplordb_schema(import_data=import_data, full_db=full_db)


@pytest.fixture()
def data_importer(sqlalchemy_session):
    importer = ImportData(sqlalchemy_session)
    yield importer


@pytest.fixture()
def db_creator(host, user, password, new_db, port):
    db_creator = CreateDatabase(host=host, user=user, password=password, connection_db='postgres', port=port)
    yield db_creator

    # Delete created database
    session = connect(host=host, port=port, user=user, password=password)
    session.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
    cur = session.cursor()
    cur.execute(f'DROP DATABASE {new_db}')
    session.commit()
    session.close()



from sqlalchemy import Column, String, ForeignKey
from xplordb.datamodel.lith import LithCode

from xplordb.sqlalchemy.base import Base


class XplordbLithCodeTable(LithCode, Base):
    """
    Define Xplordb columns for lith code definition
    """
    __tablename__ = "lithology"
    __table_args__ = {"schema": 'ref'}

    code = Column('code', String, primary_key=True)
    description = Column('description', String)
    loaded_by = Column(String, ForeignKey('ref.person.code'))



"""
Module define a lithology and associated lithology code

Module define a lithology and associated lithology code


xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/09"
__license__ = "AGPLv3"


class LithCode:

    def __init__(self, code: str, loaded_by: str, description: str = None):
        """
        Define a lithology code

        :param code: unique lithology code identifier
        :param loaded_by:  person used to load lithology code
        :param description: description of lithology code
        """
        self.code = code
        self.loaded_by = loaded_by
        if description:
            self.description = description
        else:
            self.description = code


class Lith:

    def __init__(self, lith_code: str, data_set: str, hole_id: str, loaded_by: str,
                 from_m: float, to_m: float, logged_by: str = None):
        """
        Define a lithologu

        :param lith_code: lithology code
        :param data_set: data set associated
        :param hole_id: collar used
        :param loaded_by: person loading lithology in database
        :param from_m: start of lithology from collar (meters)
        :param to_m: end of lithology from collar (meters)
        :param logged_by: person who logged lithology
        """
        self.lith_code = lith_code
        self.data_set = data_set
        self.hole_id = hole_id
        self.loaded_by = loaded_by
        self.from_m = from_m
        self.to_m = to_m
        if logged_by:
            self.logged_by = logged_by
        else:
            self.logged_by = loaded_by



from sqlalchemy import Column, String, Boolean
from sqlalchemy.orm import relationship

from xplordb.datamodel.person import Person

from xplordb.sqlalchemy.base import Base


class XplordbPersonTable(Person, Base):
    """
    Define Xplordb columns for Person definition
    """
    __tablename__ = "person"
    __table_args__ = {"schema": 'ref'}

    code = Column('code', String, primary_key=True)
    full_name = Column('person', String)
    loaded_by = Column(String)
    type = Column('type', String)
    active = Column('active', Boolean)

    datasets = relationship("XplordbDatasetTable")
    lith_codes = relationship("XplordbLithCodeTable")



"""
Module define a survey

Module define a survey


xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/09"
__license__ = "AGPLv3"


class Survey:

    def __init__(self, hole_id: str, data_set: str, loaded_by: str,
                 depth: float, dip: float, azimuth: float):
        """
        Define a survey

        :param hole_id: collar used
        :param data_set: data set associated
        :param loaded_by: person loading survey in database
        :param depth: survey depth (meters)
        :param dip: survey dip (degrees)
        :param azimuth: survey azimuth (degrees)
        """
        self.data_set = data_set
        self.hole_id = hole_id
        self.loaded_by = loaded_by
        self.depth = depth
        self.dip = dip
        self.azimuth = azimuth



from sqlalchemy import create_engine
from sqlalchemy.engine import URL
from sqlalchemy.orm import declarative_base, Session, sessionmaker

Base = declarative_base()


def create_session(host: str, port: int, database: str, user: str, password: str) -> Session:
    """
    Create a sql alchemy session with specified parameters

    :param host: host used for connection
    :param port: port used for connection
    :param database: database name used for connection
    :param user:user used for connection
    :param password:password used for connection
    """
    connection_url = URL.create(
        "postgresql+psycopg2",
        username=user,
        password=password,
        host=host,
        port=port,
        database=database
    )

    engine = create_engine(connection_url)

    xplordb_session = sessionmaker(bind=engine)
    session = xplordb_session()
    return session



# -*- coding: utf-8 -*-
import os
import re
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

requirements = (
    'psycopg2',
    'chevron',
    'sqlalchemy>=1.4,<2.0',
    'GeoAlchemy2'
)

dev_requirements = (
    'flake8',
    'pytest',
    'pytest-cov'
)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def find_version(*file_paths):
    """
    see https://github.com/pypa/sampleproject/blob/master/setup.py
    """

    with open(os.path.join(here, *file_paths), 'r') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string. "
                       "Should be at the first line of __init__.py.")


setup(
    name='xplordb',
    version=find_version('xplordb', '__init__.py'),
    description="Python module to read/write data into xplordb database",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://gitlab.com/geolandia/openlog/xplordb',
    author='Oslandia',
    author_email='geology@oslandia.com',
    license='AGPLv3',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    test_suite="tests",
    extras_require={
        'dev': dev_requirements
    },
)



"""
Python class describing an xplordb schema

xplordb schema must be created in specific order, this class define available schema
and specific schema and table import order

xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/02"
__license__ = "AGPLv3"

import os
from pathlib import Path
from typing import List

from xplordb.sql_utils import import_sql_in_directory, import_all_sql_in_directory, \
    exec_conn_psql, psql_sql_in_directory, psql_all_sql_in_directory


class Schema:
    IMPORT_ROOT = Path(os.path.dirname(os.path.abspath(__file__)))

    def __init__(self, name: str,
                 fk_tables: List[str] = None, function_after_table_import: List[str] = None,
                 extra_files: List[str] = None, fk_views: List[str] = None, full_db: bool = True):
        """
        A class used to define a xplordb schema
        :param name: schema name
        :param fk_tables: list of table .sql files that are used as foreign keys (optional, default None)
        :param function_after_table_import: list of function .sql that must be imported after table import
                                            (optional, default None)
        :param extra_files: list of .sql file that can be removed from full database (optional, default None)
        :param fk_views: list of view .sql that are used in other view (optional, default None)
        :param full_db: boolean to define if full database DDL must be imported (optional, default False)
        """
        self.name = name
        if function_after_table_import is None:
            function_after_table_import = []
        if fk_tables is None:
            fk_tables = []
        if extra_files is None:
            extra_files = []
        if fk_views is None:
            fk_views = []
        self.fk_tables = fk_tables
        self.function_after_table_import = function_after_table_import
        self.extra_files = extra_files
        self.fk_views = fk_views
        self.full_db = full_db

    @staticmethod
    def db_schema_list(full_db: bool = False):
        """
        Get xplordb schema list.
        :param full_db: boolean to define schemas for full database DDL must be created (optional, default False)
        :return: list of xplordb schema
        """
        ref = Schema('ref',
                     fk_tables=['company.sql', 'person.sql', 'data_sets.sql', 'data_source.sql', 'lab.sql',
                                'lab_o_method.sql'],
                     extra_files=['assay_result_code.sql', 'dh_survey_instrument.sql', 'elements.sql',
                                  'event.sql', 'event_code.sql', 'geomatic_declination.sql',
                                  'geomagnetic_declination.sql', 'lease.sql', 'minerals.sql',
                                  'oxidation.sql', 'preferred.sql', 'qgis_project.sql',
                                  'rl_method.sql', 'sample_class.sql', 'sample_method.sql', 'sample_type.sql',
                                  'sg_method.sql', 'strat_name.sql', 'struc.sql',
                                  'xrf_instrument.sql',
                                  'lab.sql', 'lab_method.sql', 'lab_method_code.sql', 'lab_o_method.sql',
                                  'lab_method_trigger.sql'],
                     full_db=full_db)
        dh = Schema('dh',
                    fk_tables=['collar.sql', 'surv.sql', 'details.sql', 'metadata.sql'],
                    function_after_table_import=["dh_trace.sql", "dh_planned_trace.sql"],
                    extra_files=['alteration.sql', 'core_recovery.sql', 'event.sql', 'minerals.sql',
                                 'oxidation.sql', 'petrology.sql', 'sample.sql', 'sample_image.sql',
                                 'sample_quality.sql', 'sample_weight.sql', 'sg.sql', 'struc.sql', 'vein.sql',
                                 'oxidation_trigger.sql', 'sample_quality_trigger.sql',
                                 'sample_trigger.sql'],
                    full_db=full_db)

        assay = Schema('assay',
                       extra_files=['assay.sql', 'batch.sql', 'raw.sql', 'intercepts.sql', 'flat_method.sql',
                                    'import.sql', 'flat_ppm.sql', 'structural_types.sql'],
                       fk_views=['assay.sql'],
                       function_after_table_import=['flat_method.sql', 'import.sql', 'flat_ppm.sql', "structural_types.sql"],
                       full_db=full_db)
        
        display = Schema('display',
                         fk_tables=['display_collar.sql'],
                         full_db=full_db)

        schemas = [ref, dh, assay, display]

        if full_db:
            qa = Schema('qa',
                        fk_tables=['qc_type.sql', 'sd_values.sql'])
            surf = Schema('surf')
            v = Schema('v')
            dem = Schema('dem',
                         function_after_table_import=['cap_alos_example.sql', 'alos_cap_update_example.sql'],
                         full_db=full_db)

            schemas += [qa, surf, v, dem]

        return schemas

    def create_schema(self, conn):
        print(f'importing schema: {self.name}')
        exec_conn_psql(conn, self.import_dir() / f'{self.name}.sql')

    def get_mustache_params(self) -> {}:
        params = {}
        if not self.full_db:
            params = {'geom_trace_tables': 'details'}
        return params

    def import_pre_table_functions(self, conn):
        function_list = [x for x in self.function_list()
                         if x not in self.function_after_table_import and x not in self.get_excluded_files()]
        self._import_sql_in_directory(conn, self.function_path(), function_list)

    def import_tables(self, conn):
        table_path = self.import_dir() / 'table'
        fk_tables = [x for x in self.fk_tables
                     if x not in self.get_excluded_files()]
        self._import_sql_in_directory(conn, table_path, fk_tables)
        all_exclusion = self.get_excluded_files() + fk_tables
        self._import_all_sql_in_directory(conn, table_path, all_exclusion)

    def import_post_table_functions(self, conn):
        self._import_sql_in_directory(conn, self.function_path(), self.function_after_table_import)

    def import_sequences(self, conn):
        self._import_all_sql_in_directory(conn, self.import_dir() / 'sequences')

    def import_views(self, conn):
        views_path = self.import_dir() / 'views'
        fk_views = [x for x in self.fk_views
                    if x not in self.get_excluded_files()]
        self._import_sql_in_directory(conn, views_path, fk_views)
        all_exclusion = self.get_excluded_files() + fk_views
        self._import_all_sql_in_directory(conn, views_path, all_exclusion)

    def import_fk_creation(self, conn):
        self._import_all_sql_in_directory(conn, self.import_dir() / 'fk_creation')

    def import_trigger(self, conn):
        excluded_files = self.get_excluded_files()
        self._import_all_sql_in_directory(conn, self.import_dir() / 'trigger', excluded_files)

    def import_tables_data(self, conn):
        table_path = self.import_dir() / 'populate'
        fk_tables = [x for x in self.fk_tables
                     if x not in self.get_excluded_files()]
        psql_sql_in_directory(conn, table_path, fk_tables)
        all_exclusion = self.get_excluded_files() + fk_tables
        psql_all_sql_in_directory(conn, table_path, all_exclusion)

    def get_excluded_files(self):
        excluded_files = []
        if not self.full_db:
            excluded_files = self.extra_files
        return excluded_files

    def import_dir(self):
        return self.IMPORT_ROOT / 'schema' / self.name

    def function_path(self):
        return self.import_dir() / 'functions'

    def function_list(self):
        functions_list = []
        functions_path = self.function_path()
        if functions_path.exists():
            functions_list = os.listdir(functions_path)
        return functions_list

    def _import_all_sql_in_directory(self, conn, directory: Path, excluded_files: List[str] = None):
        import_all_sql_in_directory(conn, directory, excluded_files, self.get_mustache_params())

    def _import_sql_in_directory(self, conn, directory: Path, files: List[str]):
        import_sql_in_directory(conn, directory, files, self.get_mustache_params())
